import sys
import logging
import subprocess
import pathlib
from collections import defaultdict
import json
from typing import List

import esmvalcore.experimental as esmvaltool
from esmvalcore.dataset import Dataset
from esmvalcore.local import LocalFile

from ._version import __version__

ENCODING_STDOUT = sys.stdout.encoding or "utf-8"

BTF_SBATCH = str(pathlib.Path(__file__).parent / 'bft_sbatch.sh')

# TODO allow setting different config files in the same way that
# ESMValTool does using an argument parser

# TODO setting logger default level from command line
logging.captureWarnings(True)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

logger.info("bring_from_tapes v%s", __version__)


def send_job(files: List[str]):
    """Send sbatch job to bring this files from tapes"""
    # TODO it would be nice to be able to configure the sbatch from
    # argument line
    subprocess.run(["sbatch", BTF_SBATCH] + files)


def get_size(file: LocalFile) -> int:
    """
    Get the size of a file using du with a subprocess.
    The methods from pathlib.Path and os, unfortunatelly doesn't return
    0 if the data is on tapes.
    """
    out = subprocess.run(["du", str(file)], capture_output=True)
    stdout = out.stdout.decode(ENCODING_STDOUT)
    return int(stdout.split("\t")[0])


def bring_from_tapes(recipe_file, only_test=False):
    recipe = esmvaltool.get_recipe(recipe_file)
    datasets = recipe.data.get('datasets', [])

    # Save required files in sets to avoid 'bringing' them twice
    required_files = defaultdict(set)
    # Loop over diagnostics
    for diag in recipe.data['diagnostics'].values():
        if 'variables' not in diag or diag['variables'] is None:
            # Some diagnostics may not have variables defined,
            # in that case move to the next
            continue
        # Loop over variables in a diagnostic
        for var, var_values in diag['variables'].items():
            # TODO build also auxiliary vars
            if 'short_name' not in var_values:
                # Add short_name if not given
                var_values['short_name'] = var
            # Define the target datasets (main + additional)
            var_datasets = datasets + var_values.get('additional_datasets', [])
            # Loop over target datasets
            for ds_values in var_datasets:
                # Define the dataset object
                ds_obj = Dataset(**var_values, **ds_values)
                # find the files
                ds_obj.find_files()
                # Get only LocalFiles that have size = 0 as str
                [
                    required_files[str(file.parent)].add(file.name)
                    for file in ds_obj.files
                    if isinstance(file, LocalFile)
                ]

    # Convert sets to sorted list to make it more readable
    required_files = {
        key: sorted(values)
        for key, values in required_files.items()
    }
    logger.info("Following files will be required by ESMValTool:\n%s",
                json.dumps(required_files, indent=4))
    # keep only files with size = 0
    tape_files = {
        folder: [
            file for file in files
            if get_size(pathlib.Path(folder)/file) == 0
        ]
        for folder, files in required_files.items()
    }
    # keep only folders where there are any missing file
    tape_files = {
        folder: files
        for folder, files in tape_files.items() if files
    }

    if not tape_files:
        logger.info(
            "All the required files seems to be in regular memory. "
            "There is nothing to do :)")
        return

    logger.info("Following folders have files that seem to be on tapes:\n%s",
                json.dumps(list(tape_files.keys()), indent=4))

    logger.info("Following files will be required by ESMValTool and "
                "they seem to be on tapes:\n%s",
                json.dumps(tape_files, indent=4))

    if only_test:
        # TODO change name of this parameter and set through CLI
        return

    for folder, files in tape_files.items():
        files_full = [str(pathlib.Path(folder)/file) for file in files]
        logger.info("Sending files in folder: '%s'", folder)
        logger.debug("Sending files:\n %s", files_full)
        send_job(files_full)


if __name__ == '__main__':
    bring_from_tapes(sys.argv[1])
